// $Id: PDFAction.java 10958 2009-09-29 13:19:54Z chris $

package org.faceless.pdf;

import java.util.*;

/**
 * <p>
 * An action tells the PDF viewer how to navigate around the document.
 * Actions are performed when the user clicks on a <code>PDFBookmark</code>,
 * when a {@link Form} element specifies an event, or optionally when a
 * document is opened (if the PDF's <code>setOpenAction</code> method is called).
 * </p><p>
 * The <code>PDFAction</code> class has no public constructor, just a number of
 * static methods returning a new action of the specified type.
 * </p>
 * @version $Revision: 1.3 $
 */
public final class PDFAction extends PeeredObject
{
    /**
     * Method for the {@link #formSubmit} action specifying that the form should
     * be submitted as a standard HTTP GET, in the same way as an HTML form.
     * <i><b>Note:</b> we've recently found some problems with submission via an
     * HTTP GET - specifically it doesn't work from a browser with Acrobat 5.
     * For this reason we strongly recommend using {@link #METHOD_HTTP_POST}
     * instead</i>
     */
    public static final int METHOD_HTTP_GET = 4+8;

    /**
     * Method for the {@link #formSubmit} action specifying that the form should
     * be submitted as a standard HTTP GET, in the same way as an HTML form. The
     * co-ordinates where the submit button was clicked are included in the
     * submission as <i>name.x</i> and <i>name.y</i>, where name is the name of
     * the submit button that was clicked.
     * </p><p>
     * <i><b>Note:</b> we've recently found some problems with submission via an
     * HTTP GET - specifically it doesn't work from a browser with the Acrobat 5
     * plugin. For this reason we strongly recommend using
     * {@link #METHOD_HTTP_IMAGEMAP_POST} instead</i>
     * </p>
     */
    public static final int METHOD_HTTP_IMAGEMAP_GET = 4+8+16;

    /**
     * Method for the {@link #formSubmit} action specifying that the form should
     * be submitted as a standard HTTP POST, in the same way as an HTML form
     */
    public static final int METHOD_HTTP_POST = 4;

    /**
     * Method for the {@link #formSubmit} action specifying that the form should
     * be submitted as a standard HTTP POST, in the same way as an HTML form. The
     * co-ordinates where the submit button was clicked are included in the
     * submission as <i>name.x</i> and <i>name.y</i>, where name is the name of
     * the submit button that was clicked.
     */
    public static final int METHOD_HTTP_IMAGEMAP_POST = 4+16;

    /**
     * Method for the {@link #formSubmit} action specifying that the form should
     * be submitted in the Adobe Forms Data Format.
     */
    public static final int METHOD_FDF = 0;

    /**
     * Method for the {@link #formSubmit} action specifying that the form should
     * be submitted as XML.
     * This method requires PDF 1.4 (Acrobat 5.0) or later.
     */
    public static final int METHOD_XML = 32;

    /**
     * Method for the {@link #formSubmit} action specifying that the form should
     * be submitted in the Adobe Forms Data Format, and that it should include
     * all the annotations in the document.
     * This method requires PDF 1.4 (Acrobat 5.0) or later.
     */
    public static final int METHOD_FDF_WITH_ALL_ANNOTATIONS = 128;

    /**
     * Method for the {@link #formSubmit} action specifying that the form should
     * be submitted in the Adobe Forms Data Format, and that it should include
     * any annotations that have been created by the "current user". The spec is
     * a bit vague on how the current user is determined, but we include this method
     * for completeness.
     * This method requires PDF 1.4 (Acrobat 5.0) or later.
     */
    public static final int METHOD_FDF_WITH_MY_ANNOTATIONS = 128+1024;

    /**
     * Method for the {@link #formSubmit} action specifying that the form should
     * be submitted as a PDF.
     * This method requires PDF 1.4 (Acrobat 5.0) or later.
     */
    public static final int METHOD_PDF = 256;


    final org.faceless.pdf2.PDFAction action;

    PDFAction(org.faceless.pdf2.PDFAction action)
    {
        this.action=action;
    }

    Object getPeer()
    {
        return action;
    }

    /**
     * Return a new PDFAction which jumps to the specified page in the
     * PDF document. The position and zoom on the new page is the same
     * as that of the original page.
     * @param page the page this action should jump to
     */
    public static PDFAction goTo(PDFPage page)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.goTo(page.page));
    }

    /**
     * Return a new PDFAction which jumps to the specified page in the
     * PDF document, at the specified x, y and zoom level.
     * @param page the page this action should jump to.
     * @param x the X position on the new page to scroll to, or <code>Float.NaN</code> to keep the X position unchanged.
     * @param x the Y position on the new page to scroll to, or <code>Float.NaN</code> to keep the Y position unchanged.
     * @param zoom The zoom level. A value of 0 or <code>Float.NaN</code> leaves the zoom unchanged, 1 sets the zoom to 100%, 2 to 200% and so on.
     */
    public static PDFAction goTo(PDFPage page, float x, float y, float zoom)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.goTo(page.page,x,y,zoom));
    }

    /**
     * Return a new PDFAction which jumps to the specified page in the PDF document. The
     * zoom level is set so that the entire page fits within the viewers window.
     *
     * @param page the page this action should jump to
     */
    public static PDFAction goToFit(PDFPage page)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.goToFit(page.page));
    }

    /**
     * Return a new PDFAction which jumps to the specified page in the PDF document. The
     * zoom level is set so that the width of the page fits within the viewers window.
     *
     * @param page the page this action should jump to
     * @param y if the page is too tall to fit in the windows, the Y position on the new
     * page to scroll to.
     */
    public static PDFAction goToFitWidth(PDFPage page, float y)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.goToFitWidth(page.page,y));
    }

    /**
     * Return a new PDFAction which jumps to the specified page in the PDF document. The
     * zoom level is set so that the full height of the page fits within the viewers window.
     *
     * @param page The page this action should jump to.
     * @param x if the page is too wide to fit in the windows, the X position on the new
     * page to scroll to.
     */
    public static PDFAction goToFitHeight(PDFPage page, float x)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.goToFitHeight(page.page,x));
    }

    /**
     * Return a new PDFAction which jumps to the specified page in the PDF document. The
     * zoom level is set so that the rectangle specified fills the screen
     *
     * @param page the page this action should jump to
     * @param left the left edge of the visible rectangle
     * @param bottom the bottom edge of the visible rectangle
     * @param width the width of the visible rectangle
     * @param height the height of the visible rectangle
     */
    public static PDFAction goToFitRectangle(PDFPage page, float left, float bottom, float width, float height)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.goToFitRectangle(page.page,left,bottom,width,height));
    }

    /**
     * Return a new PDFAction which jumps to the specified URL.
     * @since 1.1
     */
    public static PDFAction goToURL(java.net.URL url)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.goToURL(url));
    }

    /**
     * Return a new PDFAction which jumps to the specified URL. When used
     * from a {@link PDFAnnotation}, the annotation rectangle is treated
     * as a Server-Side ImageMap. So for example, if the user clicks 20 points
     * from the left and 30 points from the top of the link rectangle, the
     * URL that is loaded is <code>http://www.url.com/?20,30</code>.
     * @since 1.1
     */
    public static PDFAction goToURL(java.net.URL url, boolean isImageMap)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.goToURL(url,isImageMap));
    }

    /**
     * Return a new PDFAction which plays a {@link PDFSound} when activated.
     * This simpler of the two playSound methods plays the sound once at full
     * volume and with mixing turned off. See the {@link PDFSound} class
     * for more information on limitations with sound in PDF documents.
     * @since 1.1
     */
    public static PDFAction playSound(PDFSound sound)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.playSound(sound.sound));
    }

    /**
     * Return a new PDFAction which plays a {@link PDFSound} when activated.
     * @param sound the {@link PDFSound} to play.
     * @param volume the volume of the sound, from 0 (off) to 100 (maximum)
     * @param repeat whether to play the sound continuously or just once
     * @param mix whether to mix the sound in with any other playing samples, or replace them
     * @since 1.1
     */
    public static PDFAction playSound(PDFSound sound, int volume, boolean repeat, boolean mix)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.playSound(sound.sound, volume, repeat, mix));
    }

    /**
     * <p>Return a new PDFAction which runs the named action. Named actions
     * are PDF viewer dependent, so should be used with care. The following
     * actions are among those known to work in Acrobat 3 and 4 readers, and
     * loosely correspond to the equivalent actions available from the menus
     * in Acrobat.
     * </p>
     * <ul>
     * <li>Document Control<BR>
     * <tt>Open Close Print GeneralInfo FontsInfo SecurityInfo Quit</tt>
     * </li>
     * <li>Document Navigation<BR>
     * <tt>NextPage PrevPage FirstPage LastPage GoToPage</tt>
     * </li>
     * <li>Searching and Selecting<BR>
     * <tt>Find FindAgain SelectAll Copy</tt>
     * </li>
     * <li>Display<BR>
     * <tt>FullScreen ZoomTo FitPage ZoomTo ActualSize FitWidth
     * FitVisible SinglePage OneColumn TwoColumns</tt>
     * </li>
     * </ul>
     * @since 1.1
     */
    public static PDFAction named(String name)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.named(name));
    }

    /**
     * Return a new PDFAction which makes the specified {@link FormElement}
     * visible on screen.
     * @param element the form element to modify
     * @see #hideElement
     * @since 1.1.23
     */
    public static PDFAction showElement(FormElement element)
    {
	return (PDFAction)PeeredObject.getPeer(org.faceless.pdf2.PDFAction.showWidget(element.element.getAnnotation(0)));
    }

    /**
     * Return a new PDFAction which hides the specified {@link FormElement}.
     * @param element the form element to modify
     * @see #showElement
     * @since 1.1.23
     */
    public static PDFAction hideElement(FormElement element)
    {
	return (PDFAction)PeeredObject.getPeer(org.faceless.pdf2.PDFAction.hideWidget(element.element.getAnnotation(0)));
    }

    /**
     * Return a new PDFAction which submits the documents AcroForm. This should
     * be used as the <code>ONCLICK</code> event for a {@link FormButton}
     * somewhere in the form, so the form can be submitted.
     * </p><p>
     * The method parameter can be one of the various {@link #METHOD_HTTP_GET METHOD}
     * values, although please note that many of these values are only implemented in
     * Acrobat 5.0 or later.
     * </p>
     * @param url the URL to submit the form to
     * @param method the method to use when submitting the form
     * @since 1.1.23
     * @see Form
     */
    public static PDFAction formSubmit(String url, int method)
    {
	int newmethod;
	if (method==METHOD_FDF)				newmethod=org.faceless.pdf2.PDFAction.METHOD_FDF;
	else if (method==METHOD_PDF)			newmethod=org.faceless.pdf2.PDFAction.METHOD_PDF;
	else if (method==METHOD_HTTP_POST)		newmethod=org.faceless.pdf2.PDFAction.METHOD_HTTP_POST;
	else if (method==METHOD_HTTP_IMAGEMAP_POST)	newmethod=org.faceless.pdf2.PDFAction.METHOD_HTTP_IMAGEMAP_POST;
	else if (method==METHOD_XML)			newmethod=org.faceless.pdf2.PDFAction.METHOD_XML;
	else if (method==METHOD_FDF_WITH_ALL_ANNOTATIONS)	newmethod=org.faceless.pdf2.PDFAction.METHOD_FDF_WITH_ALL_ANNOTATIONS;
	else if (method==METHOD_FDF_WITH_MY_ANNOTATIONS)	newmethod=org.faceless.pdf2.PDFAction.METHOD_FDF_WITH_MY_ANNOTATIONS;
	else throw new IllegalArgumentException("No such method type");

	return new PDFAction(org.faceless.pdf2.PDFAction.formSubmit(url, newmethod));
    }

    /**
     * Return a new <code>PDFAction</code> which resets the documents AcroForm,
     * setting every field to its default values.
     * @since 1.1.23
     * @see Form
     */
    public static PDFAction formReset()
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.formReset());
    }

    /**
     * Return a new <code>PDFAction</code> which imports a Forms Data Format (FDF)
     * file into the document AcroForm, setting some or all of the forms values.
     * @param file the name of the file containing the FDF. The file is relative to
     * the location of the current document
     * @since 1.1.23
     * @see Form
     */
    public static PDFAction formImportData(String file)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.formImportData(file));
    }

    /**
     * <p>
     * Return a new <code>PDFAction</code> which executes a JavaScript action.
     * This is generally used with AcroForms, although in theory it doesn't
     * have to be. It's generally better design to create functions in the
     * document-wide JavaScript (as set by {@link PDF#setJavaScript}) and just
     * use these actions to call those methods.
     * </p>
     * @param javascript the JavaScript action to perform
     * @since 1.1.23
     * @see PDF#setJavaScript
     * @see PDF#getJavaScript
     * @see Form
     */
    public static PDFAction formJavaScript(String javascript)
    {
	return new PDFAction(org.faceless.pdf2.PDFAction.formJavaScript(javascript));
    }

    /**
     * The <code>setNext</code> method allows you to create "chains" of actions
     * which are executed whenever the first entry in the chain is run. This
     * allows more complicated processing - for example, clicking on a button
     * could play a sound and then jump to a specific page.
     * </p><p>
     * Avoiding infinite action loops is the responsibility of the programmer.
     * </p>
     * @param action the action to execute after the current one, or
     * <code>null</code> to clear it.
     * @since 1.1
     */
    public void setNext(PDFAction a)
    {
	action.setNext(a==null ? null : a.action);
    }

    /**
     * Return the action that's next in the chain after this one, or
     * <code>null</code> if no further actions are defined.
     * @since 1.1
     */
    public PDFAction getNext()
    {
	return (PDFAction)PeeredObject.getPeer(action.getNext());
    }

    /**
     * <p>
     * Return the type of action. Can be "Named:<i>value</i>" (where "value" is the
     * named action that's performed), "URL", "Sound", "GoTo",
     * "GoToFitRectangle", "GoToFit", "GoToFitHeight", "GoToFitWidth",
     * "ShowElement", "HideElement", "FormReset", "FormImportData", "FormSubmit",
     * "FormJavaScript" or one of another type of action supported by
     * the PDF specification but not by this library ("Launch", "Thread" or "Movie").
     * </p><p>
     * Note since 1.1.23 the URL value no longer includes the URL being referenced - use
     * the {@link #getURL} method to retrieve that value
     * </p>
     * @return the type of action this object represents
     * @since 1.1.12
     */
    public String getType()
    {
	return action.getType();
    }

    /**
     * For actions that refer to a page - any of the "GoTo" actions except "GoToURL" -
     * return the page the action refers to. For other types of action, returns
     * <tt>null</tt>.
     * @return the page this action refers to, or <tt>null</tt> if not applicable.
     * @since 1.1.12
     */
    public PDFPage getPage()
    {
	return (PDFPage)PeeredObject.getPeer(action.getPage());
    }

    /**
     * For "Sound" actions, return the PDFSound object this action refers to.
     * For other types of action this returns <tt>null</tt>
     * @return the sound this action refers to, or <tt>null</tt> if not applicable.
     * @since 1.1.12
     */
    public PDFSound getSound()
    {
	return (PDFSound)PeeredObject.getPeer(action.getSound());
    }

    /**
     * For <code>toggleAnnotation</code> type annotations,
     * return the annotation that is being toggled.
     * @return the annotation this action refers to, or <tt>null</tt> if not applicable.
     * @since 1.1.22
     */
    public PDFAnnotation getAnnotation()
    {
	return (PDFAnnotation)PeeredObject.getPeer(action.getAnnotation());
    }

    /**
     * For <code>JavaScript</code> type annotations, return the JavaScript
     * that is being executed. For other annotations, return <code>null</code>.
     * @return the JavaScript that is executed or <tt>null</tt> if not applicable.
     * @since 1.1.23
     */
    public String getJavaScript()
    {
	return action.getJavaScript();
    }

    /**
     * For <tt>formSubmit</tt>, <tt>goToURL</tt> and <tt>formImportData</tt>
     * actions, return the URL that is the destination of the action (or source
     * of the data). For other annotations, return <code>null</code>
     * @return the URL of the action
     * @since 1.1.23
     */
    public String getURL()
    {
	return action.getURL();
    }

    /**
     * For <tt>formSubmit</tt> actions, return the method by which the
     * form is submitted - one of the {@link #METHOD_HTTP_GET METHOD} parameters
     * defined in this class. For other action types the return is -1
     * @return the submit method for this <code>formSubmit</code> action
     * @since 1.1.23
     */
    public int getFormSubmitMethod()
    {
	return action.getFormSubmitMethod();
    }

    /**
     * Return the co-ordinates of a GoTo action. Some "GoTo" actions take
     * one or more coordinates which determine which section of the specified
     * page should be displayed. Examples of this are the
     * {@link #goToFitRectangle goToFitRectangle} method, which takes four parameters,
     * and the {@link #goToFitWidth goToFitWidth} method, which takes one. These
     * coordinates can be read by calling this method, which returns an array
     * of floats, the length of which varies depending on the type of GoTo
     * action. Specifically, "GoToFitWidth" and "GoToFitHeight" methods
     * return an array of one float, "GoToFitRectangle" will return four and
     * "GoTo" will return three. For this last one, the first number represents
     * the "X" coordinate, the second the "Y" and the third the zoom level, but
     * it's important to remember that any one of these may be unspecified -
     * in this case the value is set to <code>Float.NaN</code>.
     * @since 1.2
     */
    public float[] getGoToCoordinates()
    {
	return action.getGoToCoordinates();
    }

    public String toString()
    {
	String t = getType();
	String out = "<action type=\""+t+"\"";
	if (getJavaScript()!=null) out+=" javascript=\""+getJavaScript()+"\"";
	if (getSound()!=null) out+=" sound=\""+getSound()+"\"";
	if (getPage()!=null) out+=" page=\""+getPage()+"\"";
	if (getNext()!=null) out+=" next=\""+getNext()+"\"";
	float[] f = getGoToCoordinates();
	if (f.length>0) {
	    out+=" coords=";
	    for (int i=0;i<f.length;i++) {
	        out += (i==0 ? "\"" : ",")+f[i];
	    }
	    out+="\"";
	}
	return out+"/>";
    }
}
