// $Id: PDFAnnotation.java 10958 2009-09-29 13:19:54Z chris $

package org.faceless.pdf;

import java.util.*;
import java.awt.Color;

/**
 * <p>
 * A PDFAnnotation allows the user to interact with the generated
 * PDF document, by adding / viewing rubber-stamps and popup notes
 * or following hyperlinks. A special class of annotation, called a
 * "Widget", is also used to display {@link Form} Fields on the page.
 * </p>
 * @version $Revision: 1.4 $
 * @since 1.1
 */
public class PDFAnnotation extends PeeredObject
{
    final org.faceless.pdf2.PDFAnnotation annot;

    Object getPeer()
    {
        return annot;
    }

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when the annotation is clicked.
     * @since 1.1.23
     */
    public static final int EVENT_ONCLICK	=-1;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when the mouse pointer is moved over the annotation rectangle
     * @since 1.1.23
     */
    public static final int EVENT_ONMOUSEOVER   =0;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when the mouse pointer is moved out of the annotation rectangle
     * @since 1.1.23
     */
    public static final int EVENT_ONMOUSEOUT	=1;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when the mouse button is pressed over the annotation
     * @since 1.1.23
     */
    public static final int EVENT_ONMOUSEDOWN   =2;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when the mouse button is released over the annotation.
     * @since 1.1.23
     */
    public static final int EVENT_ONMOUSEUP	=3;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when the annotation gains focus.
     * Does not apply to buttons.
     * @since 1.1.23
     */
    public static final int EVENT_ONFOCUS	=4;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when the annotation loses focus.
     * Does not apply to buttons.
     * @since 1.1.23
     */
    public static final int EVENT_ONBLUR	=5;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when the annotation value changes.
     * Does not apply to buttons.
     * @since 1.1.23
     */
    public static final int EVENT_ONCHANGE	=6;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when a key is pressed and the annotation has focus.
     * Only applies to text boxes.
     * @since 1.1.23
     */
    public static final int EVENT_ONKEYPRESS	=7;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when a the contents of the annotation is being formatted for display.
     * Only applies to text boxes.
     * @since 1.1.23
     */
    public static final int EVENT_ONFORMAT	=8;

    /**
     * Value for {@link #setEventAction} representing the event that occurs
     * when <i>another</i> annotation changes - which makes this event a
     * little bit different to the others. You may set a JavaScript action
     * on this event which can recalculate the value of this field based on
     * the value of other fields (for example). Only applies to text boxes
     * @since 1.1.23
     */
    public static final int EVENT_ONOTHERCHANGE	=9;


    /**
     * A style of text annotation whose icon represents a "note" (the default).
     * The PDF spec doesn't define this any further.
     */
    public static final int TEXT_NOTE = 0;

    /**
     * A style of text annotation whose icon represents "comment"
     * The PDF spec doesn't define this any further.
     */
    public static final int TEXT_COMMENT = 1;

    /**
     * A style of text annotation whose icon represents "help". The PDF
     * spec doesn't define this any further.
     */
    public static final int TEXT_HELP = 2;

    /**
     * A style of text annotation whose icon represents "insert". The PDF
     * spec doesn't define this any further.
     */
    public static final int TEXT_INSERT = 3;

    /**
     * A style of text annotation whose icon represents "key". The PDF
     * spec doesn't define this any further.
     */
    public static final int TEXT_KEY = 4;

    /**
     * A style of text annotation whose icon represents "new paragraph".
     * The PDF spec doesn't define this any further.
     */
    public static final int TEXT_NEWPARAGRAPH = 5;

    /**
     * A style of text annotation whose icon represents "paragraph".
     * The PDF spec doesn't define this any further.
     */
    public static final int TEXT_PARAGRAPH = 6;

    PDFAnnotation(org.faceless.pdf2.PDFAnnotation annot)
    {
	this.annot=annot;
    }

    /**
     * <p>
     * Return a new PDFAnnotation that performs the specified action
     * when clicked. This can be used to create an HTML-style hyperlink,
     * play a sound, or perform some other {@link PDFAction}. Although this
     * method can be called directly to mark a portion of the page as a
     * hyperlink, for text hyperlinks the {@link PDFPage#beginTextLink}
     * method is probably more convenient.
     * </p><p>
     * Remember to set the co-ordinates of the annotation using the
     * <code>setRectangle</code> method.
     * </p>
     *
     * @param action the {@link PDFAction} to perform when the link is clicked
     * @param border whether to display a border around the link area
     */
    public static PDFAnnotation link(PDFAction action, boolean border)
    {
	org.faceless.pdf2.AnnotationLink link = new org.faceless.pdf2.AnnotationLink();
	link.setAction(action.action);
	if (border) {
	    org.faceless.pdf2.PDFStyle style = new org.faceless.pdf2.PDFStyle();
	    link.setStyle(style);
	}
	return new PDFAnnotation(link);
    }

    /**
     * Return a new PDFAnnotation that holds a text message. When the user clicks
     * on the annotation, a window pops up and displays the contents.
     * @param label the label to give of the popup window
     * @param contents The text to display in the window. May contain newline
     * characters for formatting.
     * @param open whether the window is open by default
     * @param type The visual appearance of the annotation when closed. May be
     * {@link #TEXT_NOTE}, {@link #TEXT_COMMENT}, {@link #TEXT_HELP},
     * {@link #TEXT_INSERT}, {@link #TEXT_KEY}, {@link #TEXT_NEWPARAGRAPH} or
     * {@link #TEXT_PARAGRAPH}
     */
    public static PDFAnnotation text(String label, String contents, boolean open, int type)
    {
	org.faceless.pdf2.AnnotationNote note = new org.faceless.pdf2.AnnotationNote();

	String newtype;
	if (type==TEXT_COMMENT) newtype = "Comment";
	else if (type==TEXT_HELP) newtype = "Help";
	else if (type==TEXT_INSERT) newtype = "Insert";
	else if (type==TEXT_KEY) newtype = "Key";
	else if (type==TEXT_NEWPARAGRAPH) newtype = "New Paragraph";
	else if (type==TEXT_PARAGRAPH) newtype = "Paragraph";
	else newtype = "Note";

	if (newtype!=null) {
	    note.setType(newtype, Color.yellow);
	}
	note.setOpen(open);
	note.setAuthor(label);
	note.setContents(contents);

	return new PDFAnnotation(note);
    }

    /**
     * Create a rubber-stamp annotation which can be added to the page.
     * Fourteen different rubber stamps are available - the parameter "type"
     * must be one of <code>Approved</code>, <code>AsIs</code>,
     * <code>Confidential</code>, <code>Departmental</code>, <code>Draft</code>,
     * <code>Experimental</code>, <code>Expired</code>, <code>Final</code>,
     * <code>ForComment</code>, <code>ForPublicRelease</code>,
     * <code>NotApproved</code>, <code>NotForPublicRelease</code>, <code>Sold</code>
     * or <code>TopSecret</code>. Any other value will result in an
     * <code>IllegalArgumengException</code> being thrown.
     * @param type the type of stamp - one of the values listed above
     * @throws IllegalArgumentException if the stamp type is unknown
     * @since 1.1.23
     */
    public static PDFAnnotation stamp(String type)
        throws IllegalArgumentException
    {
	org.faceless.pdf2.AnnotationStamp stamp = new org.faceless.pdf2.AnnotationStamp("stamp.stencil."+type, 1);
	return new PDFAnnotation(stamp);
    }

    /**
     * Set the rectangle for the annotation.
     * <ul>
     * <li>For <b>link</b> annotations, this is the rectangle that must be clicked
     * in to activate the action.</li>
     * <li>For <b>text</b> annotations, this is the rectangle describing the popup
     * window when it's opened. When it's closed, the icon is positioned at <i>x1,y1</i>.
     * </li>
     * </ul>
     * <p>
     * Note that all co-ordinates are in <b>absolute page co-ordinates</b>.
     * This means they are measured in points from the bottom-left hand
     * corner of the page, regardless of any calls to <code>rotate</code>,
     * <code>translate</code>, <code>scale</code> or <code>setCanvas</code>
     * that have been made. This restriction is part of the PDF specification.
     * </p>
     * @param x1 the X co-ordinate of the bottom-left corner of the rectangle
     * @param y1 the Y co-ordinate of the bottom-left corner of the rectangle
     * @param x2 the X co-ordinate of the top-right corner of the rectangle
     * @param y2 the Y co-ordinate of the top-right corner of the rectangle
     */
    public void setRectangle(float x1, float y1, float x2, float y2)
    {
	annot.setRectangle(x1,y1,x2,y2);
    }

    /**
     * Return the rectangle this annotation applies to on the page, or <tt>null</tt>
     * if no rectangle applies for this type of annotation.
     * @return the rectangle for this annotation, in the form of an array (<i>x1</i>, <i>y1</i>, <i>x2</i>, <i>y2</i>)
     * @since 1.1.12
     */
    public float[] getRectangle()
    {
	return annot.getRectangle();
    }

    /**
     * Get the type of annotation. Can be either "Link", "Text" or "Stamp:<i>type</i>"
     * (where <i>type</i> is the type of stamp, eg "Draft"), all of
     * which are supported by the PDF library, or one of various other annotation types
     * supported by the PDF specification but not this library.
     * @return the type of annotation
     * @since 1.1.12
     */
    public String getType()
    {
	return annot.getType();
    }

    /**
     * Return the page that this annotation is on, or <code>null</code>
     * if it a new annotation which has not yet been placed on a page
     * @since 1.1.23
     */
    public PDFPage getPage()
    {
        return (PDFPage)PeeredObject.getPeer(annot.getPage());
    }

    /**
     * Set the page for this annotation. This is useful when
     * moving annotations between documents or across pages - when
     * simply adding a new annotation to a page, this method is
     * identical to calling {@link PDFPage#addAnnotation}
     * @since 1.1.23
     */
    public void setPage(PDFPage page)
    {
	annot.setPage(page.page);
    }

    /**
     * Set whether the annotation is visible on screen. By default, all new
     * annotations are visible
     * @param screen whether the annotation should be displayed on the screen
     * @param print whether the annotation should be printed
     * @since 1.1.23
     */
    public void setVisible(boolean visible)
    {
	annot.setVisible(visible);
    }

    /**
     * Set whether the annotation is included when the document is printed. By
     * default, all new annotations are printed.
     * @param screen whether the annotation should be displayed on the screen
     * @param print whether the annotation should be printed
     * @since 1.1.23
     */
    public void setPrintable(boolean printable)
    {
	annot.setPrintable(printable);
    }

    /**
     * Returns whether the annotation is visible on the
     * screen (true) or not (false)
     * @since 1.1.23
     */
    public boolean isVisible()
    {
	return annot.isVisible();
    }

    /**
     * Returns whether the annotation is included when the document
     * is printed (true) or not (false)
     * @since 1.1.23
     */
    public boolean isPrintable()
    {
	return annot.isPrintable();
    }

    /**
     * Set the color of the annotation.
     * <ul>
     * <li>For <b>link</b> annotations, this is the color of the border around the
     * annotation (assuming the border is on)</li>
     * <li>For <b>text</b> annotations, this is color of the icon and highlights on
     * the popup window</li>
     * </ul>
     */
    public void setColor(Color c)
    {
	if (annot instanceof org.faceless.pdf2.AnnotationLink) {
	    if (c==null) {
		((org.faceless.pdf2.AnnotationLink)annot).setStyle(null);
	    } else {
		org.faceless.pdf2.PDFStyle style = new org.faceless.pdf2.PDFStyle();
		style.setLineColor(c);
		((org.faceless.pdf2.AnnotationLink)annot).setStyle(style);
	    }
	} else if (annot instanceof org.faceless.pdf2.AnnotationNote) {
	    ((org.faceless.pdf2.AnnotationNote)annot).setType(((org.faceless.pdf2.AnnotationNote)annot).getType(), c);
	}
    }

    /**
     * For a link annotation, set the action that the annotation refers to.
     * If this method is invoked on a non-link annotation it has no effect.
     * Since 1.1.23, passing in a null action removes the action completely.
     */
    public void setAction(PDFAction action)
    {
	if (annot instanceof org.faceless.pdf2.AnnotationLink) {
	    ((org.faceless.pdf2.AnnotationLink)annot).setAction(action==null ? null : action.action);
	}
    }

    /**
     * For a link annotation, get the action that the annotation refers to.
     * If this method is invoked on a non-link annotation, null is returned.
     */
    public PDFAction getAction()
    {
	if (annot instanceof org.faceless.pdf2.AnnotationLink) {
	    return (PDFAction)PeeredObject.getPeer(((org.faceless.pdf2.AnnotationLink)annot).getAction());
	} else if (annot instanceof org.faceless.pdf2.WidgetAnnotation) {
	    return (PDFAction)PeeredObject.getPeer(((org.faceless.pdf2.WidgetAnnotation)annot).getAction(org.faceless.pdf2.Event.CLICK));
	} else {
	    return null;
	}
    }

    /**
     * For a text or stamp annotation, set the contents of the annotation.
     * The string may contain newlines.
     * @since 1.1.12
     */
    public void setTextAnnotationContents(String contents)
    {
	if (annot instanceof org.faceless.pdf2.AnnotationNote) {
	    ((org.faceless.pdf2.AnnotationNote)annot).setContents(contents);
	} else if (annot instanceof org.faceless.pdf2.AnnotationStamp) {
	    ((org.faceless.pdf2.AnnotationStamp)annot).setContents(contents);
	}
    }

    /**
     * For a text or stamp annotation, get the contents of the annotation.
     * For other types of annotation, this returns <code>null</code>
     * @return the content of the text annotation or <tt>null</tt>
     * @since 1.1.12
     */
    public String getTextAnnotationContents()
    {
	if (annot instanceof org.faceless.pdf2.AnnotationNote) {
	    return ((org.faceless.pdf2.AnnotationNote)annot).getContents();
	} else if (annot instanceof org.faceless.pdf2.AnnotationStamp) {
	    return ((org.faceless.pdf2.AnnotationStamp)annot).getContents();
	} else {
	    return null;
	}
    }

    /**
     * For a text annotation, set the label of the annotation. This
     * is also used in Acrobat to represent the "author" of an
     * annotation, and as such may be used on any type of annotation.
     * The specified value may contain non-ASCII Unicode characters
     * @since 1.1.12
     */
    public void setTextAnnotationLabel(String label)
    {
	if (annot instanceof org.faceless.pdf2.AnnotationNote) {
	    ((org.faceless.pdf2.AnnotationNote)annot).setAuthor(label);
	}
    }

    /**
     * For a text annotation, get the label of the annotation. This
     * is also used in Acrobat to represent the "author" of an
     * annotation, and as such may be used on any type of annotation.
     * @return the label/author of the annotation or <tt>null</tt>
     * @since 1.1.12
     */
    public String getTextAnnotationLabel()
    {
	if (annot instanceof org.faceless.pdf2.AnnotationNote) {
	    return ((org.faceless.pdf2.AnnotationNote)annot).getAuthor();
	} else {
	    return null;
	}
    }

    /**
     * Return the date when the annotation was last modified if
     * specified, or <code>null</code> otherwise.
     * @since 1.1.23
     */
    public Calendar getLastModified()
    {
        return annot.getModifyDate();
    }

    /**
     * <p>
     * Set an action to occur when the specified event happens to
     * this annotation. These event actions are only used for annotations
     * created by form elements - those returned from {@link FormElement#getAnnotations}.
     * </p><p>
     * The event can be one of {@link #EVENT_ONCLICK},
     * {@link #EVENT_ONMOUSEOVER}, {@link #EVENT_ONMOUSEOUT},
     * {@link #EVENT_ONMOUSEDOWN}, {@link #EVENT_ONMOUSEUP},
     * {@link #EVENT_ONFOCUS}, {@link #EVENT_ONBLUR}, {@link #EVENT_ONCHANGE},
     * {@link #EVENT_ONKEYPRESS} or {@link #EVENT_ONFORMAT} - although not
     * every event type is applicable for every type of form element.
     * </p>
     * @param event the event the action applies to
     * @param action the action you wish to occur, or <code>null</code> to remove the action
     * @since 1.1.23
     * @see FormElement
     */
    public void setEventAction(int event, PDFAction action)
    {
	org.faceless.pdf2.Event newevent=null;
	if (annot instanceof org.faceless.pdf2.WidgetAnnotation) {
	    if (event==EVENT_ONCLICK) newevent=org.faceless.pdf2.Event.CLICK;
	    else if (event==EVENT_ONMOUSEOVER) newevent=org.faceless.pdf2.Event.MOUSEOVER;
	    else if (event==EVENT_ONMOUSEOUT) newevent=org.faceless.pdf2.Event.MOUSEOUT;
	    else if (event==EVENT_ONMOUSEDOWN) newevent=org.faceless.pdf2.Event.MOUSEDOWN;
	    else if (event==EVENT_ONMOUSEUP) newevent=org.faceless.pdf2.Event.MOUSEUP;
	    else if (event==EVENT_ONBLUR) newevent=org.faceless.pdf2.Event.BLUR;
	    else if (event==EVENT_ONFOCUS) newevent=org.faceless.pdf2.Event.FOCUS;
	    else if (event==EVENT_ONCHANGE) newevent=org.faceless.pdf2.Event.CHANGE;
	    else if (event==EVENT_ONFORMAT) newevent=org.faceless.pdf2.Event.FORMAT;
	    else if (event==EVENT_ONKEYPRESS) newevent=org.faceless.pdf2.Event.KEYPRESS;
	    else if (event==EVENT_ONOTHERCHANGE) newevent=org.faceless.pdf2.Event.OTHERCHANGE;
	    if (newevent!=null) {
		if (event==EVENT_ONKEYPRESS || event==EVENT_ONFORMAT || event==EVENT_ONCHANGE || event==EVENT_ONOTHERCHANGE) {
		    ((org.faceless.pdf2.WidgetAnnotation)annot).getField().setAction(newevent, action==null ? null : action.action);
		} else {
		    ((org.faceless.pdf2.WidgetAnnotation)annot).setAction(newevent, action==null ? null : action.action);
		}
	    }
	}
    }

    /**
     * Return the action that occurs when the specified event happens to
     * this annotation. This is only used with annotations returned from
     * form fields via the {@link FormElement#getAnnotations} method.
     * @param the event you want the action for - one of the events listed in {@link #setEventAction}
     * @return the action for that event, or <code>null</code> if no action is specified
     * @since 1.1.23
     * @see FormElement
     */
    public PDFAction getEventAction(int event)
    {
	if (annot instanceof org.faceless.pdf2.WidgetAnnotation) {
	    org.faceless.pdf2.Event newevent=null;
	    if (event==EVENT_ONCLICK) newevent=org.faceless.pdf2.Event.CLICK;
	    else if (event==EVENT_ONMOUSEOVER) newevent=org.faceless.pdf2.Event.MOUSEOVER;
	    else if (event==EVENT_ONMOUSEOUT) newevent=org.faceless.pdf2.Event.MOUSEOUT;
	    else if (event==EVENT_ONMOUSEDOWN) newevent=org.faceless.pdf2.Event.MOUSEDOWN;
	    else if (event==EVENT_ONMOUSEUP) newevent=org.faceless.pdf2.Event.MOUSEUP;
	    else if (event==EVENT_ONBLUR) newevent=org.faceless.pdf2.Event.BLUR;
	    else if (event==EVENT_ONFOCUS) newevent=org.faceless.pdf2.Event.FOCUS;
	    else if (event==EVENT_ONCHANGE) newevent=org.faceless.pdf2.Event.CHANGE;
	    else if (event==EVENT_ONFORMAT) newevent=org.faceless.pdf2.Event.FORMAT;
	    else if (event==EVENT_ONKEYPRESS) newevent=org.faceless.pdf2.Event.KEYPRESS;
	    else if (event==EVENT_ONOTHERCHANGE) newevent=org.faceless.pdf2.Event.OTHERCHANGE;
	    if (newevent!=null) {
		if (event==EVENT_ONKEYPRESS || event==EVENT_ONFORMAT || event==EVENT_ONCHANGE || event==EVENT_ONOTHERCHANGE) {
		    return (PDFAction)PeeredObject.getPeer(((org.faceless.pdf2.WidgetAnnotation)annot).getField().getAction(newevent));
		} else {
		    return (PDFAction)PeeredObject.getPeer(((org.faceless.pdf2.WidgetAnnotation)annot).getAction(newevent));
		}
	    } else {
		return null;
	    }
	} else {
	    return null;
	}
    }

    public String toString()
    {
        String out = "<annotation type=\""+getType()+"\" rect=\"";
	float[] f = getRectangle();
	out+=f[0]+","+f[1]+","+f[2]+","+f[3]+"\"";
	if (getType().equals("Text")) {
	    out+=" label=\""+getTextAnnotationLabel()+"\" content=\""+getTextAnnotationContents()+"\"";
	} else if (getType().equals("Widget")) {
	    String n = ((org.faceless.pdf2.WidgetAnnotation)annot).getValue();
	    if (n!=null) {
	        out+=" value=\""+n+"\"";
	    }
	}
	if (!isVisible()) out+=" visible=\"false\"";
	if (!isPrintable()) out+=" printable=\"false\"";
	if (getAction()!=null) {
	    out+=" action=\""+getAction()+"\"";
	}
	String[] triggers = { "E","X","D","U","Fo","Bl","V","K","F","C" };
	for (int i=0;i<triggers.length;i++) {
	    final String[] names = { "onMouseOver", "onMouseOut", "onMouseDown", "onMouseUp", "onFocus", "onBlur", "onChange", "onKeyPress", "onFormat", "onOtherChange" };
	    PDFAction a = getEventAction(i);
	    if (a!=null) {
	       out+=" "+names[i]+"=\""+a+"\"";
	    }
	}
	return out+"/>";
    }

}
