// $Id: MultiPageDocumentViewport.java 16047 2012-07-16 09:29:03Z mike $

package org.faceless.pdf2.viewer2;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.beans.*;
import java.util.List;
import javax.swing.*;
import org.faceless.pdf2.*;

/**
 * A {@link DocumentViewport} that displays all the pages of a PDF in one column
 * (or row) in the viewport. Navigating between pages is done by scrolling the viewport.
 *
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @see DocumentPanel
 * @since 2.10.4
 */
public class MultiPageDocumentViewport extends DocumentViewport {

    private JScrollPane scrollPane;
    private View view;
    private float zoom;

    private Listener listener;
    private RenderingHints hints;
    private ArrayList pagePanelListeners;
    private ArrayList pagePanelInteractionListeners;
    private PDF pdf;
    private PDFParser parser;

    /**
     * Constructs a new MultiPageDocumentViewport with a vertical
     * orientation.
     */
    public MultiPageDocumentViewport() {
        this(Adjustable.VERTICAL);
    }

    /**
     * Constructs a new multi-page document viewport with the specified
     * orientation.
     * @param orientation Adjustable.VERTICAL or Adjustable.HORIZONTAL
     */
    public MultiPageDocumentViewport(int orientation) {
        view = this.new View(orientation);
        setRenderingHints(getDefaultRenderingHints());
        setLayout(new BorderLayout());
        scrollPane = new JScrollPane(view) {
            protected JViewport createViewport() {
                return new JViewport() {
                    public void scrollRectToVisible(Rectangle r) {
                        // NB eat this event to prevent caret from scrolling viewport
                    }
                };
            }
        };
        add(scrollPane, BorderLayout.CENTER);
        zoom = 1;

        // event management
        listener = this.new Listener();
        addComponentListener(listener);
        scrollPane.getHorizontalScrollBar().addAdjustmentListener(listener);
        scrollPane.getVerticalScrollBar().addAdjustmentListener(listener);

        pagePanelListeners = new ArrayList();
        pagePanelInteractionListeners = new ArrayList();

        Util.fixScrollPaneKeyBindings(scrollPane);
    }

    public void setDocumentPanel(DocumentPanel panel) {
        super.setDocumentPanel(panel);
        if (panel != null) {
            setPDF(panel.getPDF());
            parser = panel.getParser();
            view.reorder();
            scrollPane.getHorizontalScrollBar().setUnitIncrement(mouseWheelUnit);
            scrollPane.getVerticalScrollBar().setUnitIncrement(mouseWheelUnit);
        } else {
            pdf = null;
            parser = null;
            view.dispose();
        }
    }

    public boolean isDraggable() {
        return view.isDraggable();
    }

    public void setDraggable(boolean draggable) {
        view.setDraggable(draggable);
    }

    public void setPage(PDFPage page, double x, double y, double zoom) {
        setPDF(page.getPDF());
        setZoom((float) zoom);
        Rectangle2D rect = PagePanel.getFullPageView(page);
        ensureVisible(page, x, rect.getHeight() - y, false);
    }

    public void ensureVisible(PDFPage page, double x, double y) {
        ensureVisible(page, x, y, true);
    }

    /**
     * @param x x coordinate in PDF units
     * @param y y coordinate in PDF units
     * @param centre if true, centre this location in the viewport,
     * otherwise cause it to be displayed at the top left of the viewport.
     */
    private void ensureVisible(PDFPage page, double x, double y, boolean centre) {
        Rectangle pageRect = view.getPageRectangle(page.getPageNumber() - 1);
        if (pageRect != null) {
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            int xoffset = 0, yoffset = 0;
            if (!Double.isNaN(x)) {
                xoffset = pointsToPixels((float) x);
            }
            if (!Double.isNaN(y)) {
                yoffset = pointsToPixels((float) y);
                // convert to AWT axis space
                yoffset = pageRect.height - yoffset;
            }
            xoffset = pageRect.x + xoffset;
            yoffset = pageRect.y + yoffset;
            if (centre) {
                xoffset -= hsb.getVisibleAmount() / 2;
                yoffset -= vsb.getVisibleAmount() / 2;
            }
            if (getPagePanel() != null && getPagePanel().getClip() != null) {
                hsb.setValue(xoffset);  // Don't want to smooth scroll if we're clipping
                vsb.setValue(yoffset);  // the page rectangles.
            } else {
                smoothScroll(xoffset, yoffset, hsb, vsb);
            }
        }
    }

    public boolean isPageVisible(PDFPage page, double x, double y) {
        Rectangle pageRect = view.getPageRectangle(page.getPageNumber() - 1);
        if (pageRect != null) {
            Rectangle target = (Rectangle) pageRect.clone();
            float dpi = getZoom() * Util.getScreenResolution(this);
            if (!Double.isNaN(x)) {
                int xoffset = Math.round((float) x * (dpi / 72));
                target.x += xoffset;
                target.width = 8;
            }
            if (!Double.isNaN(y)) {
                int yoffset = Math.round((float) y * (dpi / 72));
                // convert to PDF axis space
                yoffset = pageRect.height - yoffset;
                target.y += yoffset;
                target.height = 8;
            }
            return getClip().intersects(target);
        }
        return false;
    }

    public JComponent getView() {
        return (JComponent)scrollPane.getViewport().getView();
    }

    public float getZoom() {
        if (zoom==0) {
            zoom = getTargetZoom(getZoomMode(), getPage());
        }
        return zoom;
    }

    public void setZoom(float zoom) {
        if (zoom == zoom && zoom > 0.01 && Math.abs(zoom - this.zoom) > 0.01) {
            final JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            final JScrollBar vsb = scrollPane.getVerticalScrollBar();
            final float hv = (float)hsb.getValue() / hsb.getMaximum();
            final float vv = (float)vsb.getValue() / vsb.getMaximum();
            this.zoom = zoom;
            view.redimension();

            // set up what we want to do but don't execute until after invalidate()/validate() has completed
            // as that is necessary to set the scrollbar's new maximum values, so if we try and
            // set a value here that's > maximum (e.g. zoom in on last page of document) then it will fail
            final Dimension viewsize = view.getPreferredSize();
            Runnable adjustScroll = new Runnable() {
                public void run() {
                    listener.setEnabled(false); // don't send adjustment events
                    hsb.setValue(Math.round(viewsize.width * hv));
                    vsb.setValue(Math.round(viewsize.height * vv));
                    listener.setEnabled(true);
                    view.revalidate();
                    view.repaint();
                }
            };
            if (SwingUtilities.isEventDispatchThread()) {
                adjustScroll.run();
            } else {
                SwingUtilities.invokeLater(adjustScroll);
            }
        }
    }

    private int pointsToPixels(float points) {
        return Math.round(points * (getZoom() * Util.getScreenResolution(this) / 72));
    }

    private float pixelsToPoints(int pixels) {
        return (float)pixels / (getZoom() * Util.getScreenResolution(this) / 72);
    }

    public Dimension getViewportSize() {
        return scrollPane.getViewport().getSize();
    }

    public PagePanel getPagePanel() {
        return view.getPagePanel();
    }

    public Collection getPagePanels() {
        return view.getPagePanels();
    }

    public PDFPage getRenderingPage() {
        return getPage();
    }

    private void setPDF(PDF pdf) {
        if (this.pdf != pdf) {    // Only happens once unless not part of a DocumentPanel
            this.pdf = pdf;
        }
    }

    private Rectangle getClip() {
        JScrollBar hsb = scrollPane.getHorizontalScrollBar();
        JScrollBar vsb = scrollPane.getVerticalScrollBar();
        return new Rectangle(hsb.getValue(), vsb.getValue(), hsb.getVisibleAmount(), vsb.getVisibleAmount());
    }

    public Adjustable getAdjustable(int position) {
        switch (position) {
          case Adjustable.HORIZONTAL:
            return scrollPane.getHorizontalScrollBar();
          case Adjustable.VERTICAL:
            return scrollPane.getVerticalScrollBar();
          default:
            throw new IllegalArgumentException(Integer.toString(position));
        }
    }

    public void setAdjustableValues(int horizontal, int vertical) {
        scrollPane.getHorizontalScrollBar().setValue(horizontal);
        scrollPane.getVerticalScrollBar().setValue(vertical);
        view.revalidate();
        view.repaint();
    }

    public void setRenderingHints(RenderingHints hints) {
        this.hints = hints;
        for (Iterator i = view.getPagePanels().iterator();i.hasNext();) {
            ((PagePanel)i.next()).setRenderingHints(hints);
        }
    }

    public void addPagePanelListener(PagePanelListener listener) {
        pagePanelListeners.add(listener);
    }

    public void removePagePanelListener(PagePanelListener listener) {
        pagePanelListeners.remove(listener);
    }

    public void addPagePanelInteractionListener(PagePanelInteractionListener listener) {
        pagePanelInteractionListeners.add(listener);
    }

    public void removePagePanelInteractionListener(PagePanelInteractionListener listener) {
        pagePanelInteractionListeners.remove(listener);
    }

    public void setZoomMode(int zoommode) {
        super.setZoomMode(zoommode);
        if (getPage() != null) {
            setZoom(getTargetZoom(zoommode, getPage()));
        }
    }

    public float getTargetZoom(int zoommode, PDFPage page) {
        float outzoom = zoom;
        if (page!=null) {
            Dimension pps = view.getLargestPageSize();
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            Dimension size = new Dimension(hsb.getVisibleAmount(), vsb.getVisibleAmount());
            switch (zoommode) {
              case ZOOM_FITWIDTH:
                size.width -= vsb.getWidth();
                outzoom = (float)size.width / pps.width;
                break;
              case ZOOM_FITHEIGHT:
                size.height -= hsb.getHeight();
                outzoom = (float)size.height / pps.height;
                break;
              case ZOOM_FIT:
                size.width -= vsb.getWidth();
                size.height -= hsb.getHeight();
                float zw = (float)size.width / pps.width;
                float zh = (float)size.height / pps.height;
                outzoom = zw < zh ? zw : zh;
                break;
              default:
                outzoom = zoom==0 ? 1 : zoom;
            }
        }
        return outzoom;
    }

    public void propertyChange(PropertyChangeEvent event) {
        String name = event.getPropertyName();
        if ((name.equals("pages") && event.getSource() == pdf) || ((name.endsWith("Box") || name.equals("orientation")) && event.getSource() instanceof PDFPage)) {
            smoothScroll(0, 0, null, null);
            view.reorder();
            view.redimension();
            view.revalidate();
            view.repaint();
        }
    }

    private class View extends JPanel implements PagePanelListener, PagePanelInteractionListener, MouseListener, MouseMotionListener {

        static final int VIEW_MARGIN = 4;
        static final int INTER_PAGE_MARGIN = 10;

        private final boolean vertical;
        private Dimension viewSize, maxPageSize;
        private PagePanel[] pagePanels;         // Array of 0..n of panels - usually null
        private Rectangle[] pageRects;          // Array of 0..n of page rectangles, always set
        private Rectangle2D.Float[] fullPageRects;
        private BitSet visiblePages;            // Bitset showing which pages are visible
        private PagePanel mainPagePanel;

        private boolean draggable, pageRectsCentered;
        private Point mouseDownEvent;
        private Point mouseDownScroll;

        View(int orientation) {
            if (orientation != Adjustable.VERTICAL && orientation != Adjustable.HORIZONTAL) {
                throw new IllegalArgumentException("Invalid orientation: "+orientation);
            }
            this.vertical = orientation == Adjustable.VERTICAL;
            setLayout(null); // we will handle layout
            setFocusable(false);
            addMouseListener(this);
            addMouseMotionListener(this);
        }

        /**
         * Called when the document (maybe) has had its pages reordered.
         * Creates "pagePanels" and "visible"
         */
        void reorder() {
            if (pdf == null) {
                if (pagePanels != null) {
                    for (int i=0;i<pagePanels.length;i++) {
                        removePagePanel(i);
                    }
                    pagePanels = null;
                    visiblePages = null;
                }
                fullPageRects = null;
            } else {
                List pages = pdf.getPages();
                int numpages = pages.size();
                PagePanel[] newPagePanels = new PagePanel[numpages];
                BitSet newVisiblePages = new BitSet();
                if (pagePanels != null) {
                    for (int i=0;i<pagePanels.length;i++) {
                        PagePanel panel = pagePanels[i];
                        if (panel != null) {
                            int index = pages.indexOf(panel.getPage());
                            if (index < 0) {
                                removePagePanel(i);
                            } else {
                                newPagePanels[index] = panel;
                                newVisiblePages.set(index);
                            }
                        }
                    }
                }
                visiblePages = newVisiblePages;
                pagePanels = newPagePanels;

                fullPageRects = new Rectangle2D.Float[numpages];
                for (int i=0;i<fullPageRects.length;i++) {
                    fullPageRects[i] = (Rectangle2D.Float)PagePanel.getFullPageView(pdf.getPage(i));
                    if (fullPageRects[i] == null) {
                        throw new Error();
                    }
                }
            }
            revalidate();
            repaint();
        }

        /**
         * Called when the view or the document has been resized
         */
        void redimension() {
            viewSize = maxPageSize = null;
            pageRects = null;
        }

        PagePanel getPagePanel() {
            return mainPagePanel;
        }

        public Dimension getPreferredSize() {
            if (viewSize == null) {
                if (pagePanels == null) {
                    reorder();
                }
                if (pdf != null) {
                    int numpages = pdf.getNumberOfPages();
                    viewSize = new Dimension(0, 0);
                    maxPageSize = new Dimension(0, 0);
                    pageRects = new Rectangle[numpages];
                    int totwidth = 0, totheight = 0;
                    float maxwidth = 0, maxheight = 0;
                    int pagex = VIEW_MARGIN;
                    int pagey = VIEW_MARGIN;
                    pageRectsCentered = false;

                    for (int i = 0; i < numpages; i++) {
                        Rectangle2D.Float pageSize = fullPageRects[i];
                        int pagewidth = pointsToPixels(pageSize.width);
                        int pageheight = pointsToPixels(pageSize.height);
                        if (pageSize.width > maxwidth) {
                            maxwidth = pageSize.width;
                        }
                        if (pageSize.height > maxPageSize.height) {
                            maxheight = pageSize.height;
                        }
                        totwidth += pageSize.width;
                        totheight += pageSize.height;
                        Rectangle pageRect = new Rectangle(pagex, pagey, pagewidth, pageheight);
                        pageRects[i] = pageRect;
                        if (vertical) {
                            pagey += pageheight + INTER_PAGE_MARGIN;
                        } else {
                            pagex += pagewidth + INTER_PAGE_MARGIN;
                        }
                    }

                    if (vertical) { 
                        viewSize.width = pointsToPixels(maxwidth) + VIEW_MARGIN*2;
                        viewSize.height = pointsToPixels(totheight) + VIEW_MARGIN*2 + (numpages-1)*INTER_PAGE_MARGIN; 
                    } else {
                        viewSize.width = pointsToPixels(totwidth) + VIEW_MARGIN*2 + (numpages-1)*INTER_PAGE_MARGIN; 
                        viewSize.height = pointsToPixels(maxheight) + VIEW_MARGIN*2;
                    }

                    int dpi = Util.getScreenResolution(MultiPageDocumentViewport.this);
                    maxPageSize = new Dimension(Math.round(maxwidth * dpi / 72), Math.round(maxheight * dpi / 72));
                } else {
                    viewSize = new Dimension(0, 0);
                    maxPageSize = new Dimension(0, 0);
                    pageRects = new Rectangle[0];
                }
            }
            return viewSize;
        }

        /**
         * Return a BitSet showing which pages should be visible based
         * on the current scrollbar position
         */
        BitSet getVisiblePages() {
            if (pdf == null) {
                return new BitSet();
            }
            Dimension viewport = getPreferredSize();
            Rectangle clip = new Rectangle();
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            clip.x = hsb.getValue();
            clip.y = vsb.getValue();
            clip.width = hsb.getVisibleAmount();
            clip.height = vsb.getVisibleAmount();
            BitSet b = new BitSet(pageRects.length);
            for (int i=0;i<pageRects.length;i++) {
                Rectangle r = pageRects[i];
                if (vertical) { 
                    if (r.y < clip.y + clip.height && r.y + r.height > clip.y) {
                        b.set(i);
                    }
                } else {
                    if (r.x < clip.x + clip.width && r.x + r.width > clip.x) {
                        b.set(i);
                    }
                }
            }
            return b;
        }

        Dimension getLargestPageSize() { // NB assumes a zoom of 1.0
            getPreferredSize();
            return maxPageSize;
        }

        public void doLayout() {
            Rectangle clip = new Rectangle();
            JScrollBar hsb = scrollPane.getHorizontalScrollBar();
            JScrollBar vsb = scrollPane.getVerticalScrollBar();
            clip.x = hsb.getValue();
            clip.y = vsb.getValue();
            clip.width = hsb.getVisibleAmount();
            clip.height = vsb.getVisibleAmount();
            float dpi = getZoom() * Util.getScreenResolution(MultiPageDocumentViewport.this);

            BitSet newVisiblePages = getVisiblePages();

            // 1. If page rectangles haven't been centered, do it here and
            // position pagepanels as well.
            if (!pageRectsCentered) {       // If pageRects haven't been centered yet, do it
                Dimension size = getSize();
                Dimension viewport = getPreferredSize();
                int cx = Math.max((size.width - viewport.width) / 2, 0);
                int cy = Math.max((size.height - viewport.height) / 2, 0);
                for (int i=0;i<pageRects.length;i++) {
                    Rectangle rect = pageRects[i];
                    if (vertical) {
                        rect.x = cx + (viewport.width - rect.width)/2;
                    } else {
                        rect.y = cy + (viewport.height - rect.height)/2;
                    }
                    if (pagePanels[i] != null) {
                        pagePanels[i].setBounds(rect);
                    }
                }
                pageRectsCentered = true;
            }

            // 2. Check for changes in the set of visible pages.
            visiblePages.xor(newVisiblePages);
//            System.out.println("doLayout: VISIBLE="+newVisiblePages+" CHANGE="+visiblePages);
            if (!visiblePages.isEmpty()) {
                for (int i = visiblePages.nextSetBit(0); i >= 0; i = visiblePages.nextSetBit(i+1)) {
                    if (newVisiblePages.get(i)) {     // Page i was added
                        pagePanels[i] = addPagePanel(i, pdf.getPage(i), clip, pageRects[i]);
                    } else {                    // Page i was removed
                        removePagePanel(i);
                    }
                }
            }

            // 3. Of visible pages, position and determine which one is our "main" page.
            // Main page is the first fully visble page, or the page with most area displayed
            int newmainpage = -1, maxvispixels = 0;
            for (int i = newVisiblePages.nextSetBit(0); i >= 0; i = newVisiblePages.nextSetBit(i+1)) {
                repaintPanel(i, null, clip);
                Rectangle visible = clip.intersection(pageRects[i]);
                if (maxvispixels < Integer.MAX_VALUE) {
                    if (visible.width == pageRects[i].width && visible.height == pageRects[i].height) {
                        newmainpage = i;
                        maxvispixels = Integer.MAX_VALUE;
                    }
                    int vispixels = visible.width * visible.height;
                    if (vispixels > maxvispixels) {
                        maxvispixels = vispixels;
                        newmainpage = i;
                    }
                }
            }

            // Maybe fire pageChanged and pagePositionChanged events
            final DocumentPanel docpanel = getDocumentPanel();
            if (docpanel != null) {
                if (pagePanels[newmainpage] != mainPagePanel) {
                    mainPagePanel = pagePanels[newmainpage];
                    if (mainPagePanel.getPage() == null) {
                        // If main page hasn't completed rendering (possibly, but not always because
                        // we just created it) don't fire pageChanged until page redrawn, otherwise
                        // DocumentPanel.getPage() == null
                        final int x = newmainpage;
                        mainPagePanel.addPagePanelListener(new PagePanelListener() {
                            public void pageUpdated(PagePanelEvent event) {
                                if (event.getType() == "redrawn") {
                                    event.getPagePanel().removePagePanelListener(this);
                                    // Check we're still main page before firing event, just in case we've moved on
                                    if (event.getPagePanel() == mainPagePanel) {
                                        docpanel.raiseDocumentPanelEvent(DocumentPanelEvent.createPageChanged(docpanel));
                                    }
                                }
                            }
                        });
                    } else {
                        docpanel.raiseDocumentPanelEvent(DocumentPanelEvent.createPageChanged(docpanel));
                    }
                }
            }
            visiblePages = newVisiblePages;
        }

        /**
         * @page the page to display
         * @viewClip the clip rectangle
         * @pr the page rectangle
         */
        private PagePanel addPagePanel(int i, PDFPage page, Rectangle viewClip, Rectangle pageRect) {
            PagePanel pagePanel = new PagePanel();
            pagePanel.setViewport(MultiPageDocumentViewport.this);
            pagePanel.setParser(parser);
            if (hints != null) {
                pagePanel.setRenderingHints(hints);
            }
            pagePanel.addPagePanelListener(this);
            pagePanel.addPagePanelInteractionListener(this);
            pagePanel.addMouseListener(this);
            pagePanel.addMouseMotionListener(this);
            pagePanel.setBounds(pageRect);
            pagePanels[i] = pagePanel;
            add(pagePanel);
            repaintPanel(i, page, viewClip);
            return pagePanel;
        }

        private void repaintPanel(int i, PDFPage page, Rectangle viewClip) {
            if (page == null) {
                page = pagePanels[i].getPage(); // Quickest
                if (page == null) {
                    page = pdf.getPage(i);      // Slowest
                }
            }
            Rectangle pageClip = pageRects[i].intersection(viewClip);
            pageClip.x -= pageRects[i].x;
            pageClip.y -= pageRects[i].y;
            float dpi = getZoom() * Util.getScreenResolution(MultiPageDocumentViewport.this);
            // Next method called on every adjust, but PagePanel filters out unnecessary repaints
            pagePanels[i].setPage(page, fullPageRects[i], dpi, pageClip);
        }

        private void removePagePanel(int i) {
            PagePanel pagePanel = pagePanels[i];
            if (pagePanel != null) {
                pagePanel.removePagePanelListener(this);
                for (Iterator j = pagePanelListeners.iterator(); j.hasNext(); ) {
                    pagePanel.removePagePanelListener((PagePanelListener) j.next());
                }
                for (Iterator j = pagePanelInteractionListeners.iterator(); j.hasNext(); ) {
                    pagePanel.removePagePanelInteractionListener((PagePanelInteractionListener) j.next());
                }
                remove(pagePanel);
                pagePanel.setParser(null);
                pagePanels[i] = null;
            }
        }

        public void paintComponent(Graphics gg) {
            Graphics2D g = (Graphics2D) gg;
            super.paintComponent(g);
            if (zoom != 0) {
                // Paint page borders
                Color dark = new Color(0xFF666666, false);
                Color medium = new Color(0x80666666, true);
                Color light = new Color(0x30666666, true);
                if (visiblePages != null && pageRects != null) {
                    for (int i = visiblePages.nextSetBit(0); i >= 0; i = visiblePages.nextSetBit(i+1)) {
                        Rectangle r = pageRects[i];
                        int x1 = r.x - 1, y1 = r.y - 1;
                        int x2 = r.x + r.width + 1, y2 = r.y + r.height + 1;
                        g.setColor(dark);
                        g.drawRect(x1, y1, r.width + 2, r.height + 2);
                        y2++;
                        g.setColor(medium);
                        g.drawLine(x1 + 1, y2, x2 - 1, y2);
                        y2++;
                        g.setColor(light);
                        g.drawLine(x1 + 2, y2, x2 - 2, y2);
                    }
                }
            }
        }
    
        public void pageUpdated(PagePanelEvent event) {
            if (event.getType() == "redrawn") {
                PagePanel panel = event.getPagePanel();
                panel.validate();
                repaint();
            }
            List l = new ArrayList(pagePanelListeners);
            for (Iterator i = l.iterator(); i.hasNext(); ) {
                ((PagePanelListener) i.next()).pageUpdated(event);
            }
        }

        public void pageAction(PagePanelInteractionEvent event) {
            List c = new ArrayList(pagePanelInteractionListeners);
            for (Iterator i = c.iterator(); i.hasNext(); ) {
                ((PagePanelInteractionListener) i.next()).pageAction(event);
            }
        }

        boolean isDraggable() {
            return draggable;
        }

        void setDraggable(boolean draggable) {
            this.draggable = draggable;
            setCursor(draggable ? CURSOR_GRAB : null);
        }

        public void mouseClicked(MouseEvent event) {}
        public void mouseEntered(MouseEvent event) {}
        public void mouseExited(MouseEvent event) {}
        public void mouseMoved(MouseEvent event) {}

        public void mousePressed(MouseEvent event) {
            if (draggable) {
                smoothScroll(0, 0, null, null);
                Component source = (Component) event.getSource();
                if (source == this) {
                    mouseDownEvent = event.getPoint();
                } else {
                    mouseDownEvent = new Point(event.getX() + source.getX(), event.getY() + source.getY());
                }
                Adjustable sbh = getAdjustable(Adjustable.HORIZONTAL);
                Adjustable sbv = getAdjustable(Adjustable.VERTICAL);
                mouseDownScroll = new Point(sbh.getValue(), sbv.getValue());
                setCursor(CURSOR_GRABBING);
            }
        }

        public void mouseReleased(MouseEvent event) {
            if (draggable) {
                mouseDownEvent = null;
                mouseDownScroll = null;
                setCursor(CURSOR_GRAB);
            }
        }

        public void mouseDragged(MouseEvent event) {
            if (draggable && mouseDownEvent != null) {
                smoothScroll(0, 0, null, null);
                Point point;
                Component source = (Component) event.getSource();
                if (source == this) {
                    point = event.getPoint();
                } else {
                    point = new Point(event.getX() + source.getX(), event.getY() + source.getY());
                }
                int dx = point.x - mouseDownEvent.x;
                int dy = point.y - mouseDownEvent.y;
                Adjustable sbh = getAdjustable(Adjustable.HORIZONTAL);
                Adjustable sbv = getAdjustable(Adjustable.VERTICAL);
                int sbx = sbh.getValue();
                int sby = sbv.getValue();
                setAdjustableValues(mouseDownScroll.x - dx, mouseDownScroll.y - dy);
                int sdx = sbh.getValue() - sbx;
                int sdy = sbv.getValue() - sby;
                mouseDownScroll.x += sdx;
                mouseDownScroll.y += sdy;
            }
        }

        void dispose() {
            if (pagePanels != null) {
                for (int i=0;i<pagePanels.length;i++) {
                    removePagePanel(i);
                }
            }
            pagePanels = null;
            fullPageRects = null;
            pageRects = null;
            visiblePages = null;
        }

        Collection getPagePanels() {
            List l = new ArrayList();
            if (pagePanels != null) {
                for (int i=0;i<pagePanels.length;i++) {
                    if (pagePanels[i] != null) {
                        l.add(pagePanels[i]);
                    }
                }
            }
            return Collections.unmodifiableList(l);
        }

        Rectangle getPageRectangle(int i) {
            getPreferredSize();
            return pageRects[i];
        }
    }

    private class Listener implements ComponentListener, AdjustmentListener {

        private boolean enabled = true;

        void setEnabled(boolean enabled) {
            this.enabled = enabled;
        }

        public void componentResized(ComponentEvent e) {
            int zoommode = getZoomMode();
            switch (zoommode) {
              case ZOOM_FITWIDTH:
              case ZOOM_FITHEIGHT:
              case ZOOM_FIT:
                setZoom(getTargetZoom(zoommode, null));
                break;
            }
            smoothScroll(0, 0, null, null);
        }

        public void componentMoved(ComponentEvent e) {
        }

        public void componentShown(ComponentEvent e) {
        }

        public void componentHidden(ComponentEvent e) {
        }

        public void adjustmentValueChanged(AdjustmentEvent e) {
            if (enabled) {
                view.revalidate();
                view.repaint();
                DocumentPanel docpanel = getDocumentPanel();
                if (docpanel != null && !isSmoothScrolling()) {
                    docpanel.raiseDocumentPanelEvent(DocumentPanelEvent.createPagePositionChanged(docpanel));
                }
            }
        }

    }

}
